@@ -4,17 +4,20 @@
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
+#include <sys/types.h>
+#include <sys/file.h>
+#include <sys/stat.h>
+#include <fcntl.h>
 #include <stdarg.h>
+#include <assert.h>
+#include <errno.h>
+#include <unistd.h>
+
 
 #ifdef USE_MPI
 #include <mpi.h>
 #endif
 
-#ifndef HAVE_ASPRINTF
-extern int asprintf(char **ret, const char *format, ...);
-/* extern int vasprintf(char **ret, const char *format, va_list args); */
-#endif /* HAVE_ASPRINTF */
-
 #include "GTG.h"
 #include "GTGPaje_Basic.h"
 #include "GTGPaje.h"
@@ -25,19 +28,22 @@ extern int asprintf(char **ret, const char *format, ...);
 
 
 extern int _is_paje_header_written;
-void pajeInitHeaderData( int fmt );
+void pajeInitHeaderData( int fmt, gtg_flag_t paje_flags);
 void pajeWriteHeader( FILE *file );
 
 /* set in GTGBasic1_c.c
  * verbose !=0 means debugging mode
  */
 extern int verbose;
+/* Max length of the names built */
+#define BUFFSIZE 1024
 
-/* File where each proc writes state/event/var/link changes. */
-static FILE*  procFile;
 /* File where each proc writes the header and the definitions of the data. */
-static FILE*  headFile;
-/* Root name */
+static FILE*  headFile=NULL;
+/* name of the root file*/
+static char  headFilename[BUFFSIZE];
+
+/* base name */
 static char*  filename;
 /* mpi rank */
 static int rank;
@@ -45,8 +51,6 @@ static int rank;
 static char* nameTmp;
 /* Flags that should be used. */
 static gtg_flag_t paje_flags;
-/* Max length of the names built */
-#define BUFFSIZE 200
 
 
 /* in non-buffered mode, flush a file */
@@ -75,6 +79,90 @@ typedef struct doubleLinkList_t{
     line_t current;                /* Current line     */
 }doubleLinkList_t;
 
+struct container_file{
+  char container_name[1024];
+  char file_name[1024];
+  FILE* file;
+  int closed;
+  struct container_file *next;
+};
+
+/* File where each proc writes state/event/var/link changes. */
+static struct container_file *procFile=NULL; /* TODO: don't hardcode the size */
+static int nb_containers = 0;
+static int nb_allocated_containers = 0;
+
+/* returns 1 if str is null (or contains "0" or "(null)") */
+static int null_string(const char*str) {
+  return (! str                 ||
+	  strcmp(str, "0") == 0 ||
+	  strcmp(str, "")  == 0 ||
+	  strcmp(str, "(null)") == 0);
+}
+
+#define GET_FILE_DESCRIPTOR(f, cont)					\
+  do {									\
+    struct container_file* p_cont = __paje_find_container(cont);	\
+    if(! p_cont) {							\
+      p_cont = __paje_create_container_file(cont);			\
+    }									\
+    f = p_cont->file;							\
+  } while(0)
+
+
+/* find the container_file structure that corresponds to cont */
+static struct container_file* __paje_find_container(const char* cont) {
+  int i;
+  cont = cont? cont : "";
+  for(i=0; i<nb_containers; i++) {
+    if(!strcmp(cont, procFile[i].container_name))
+      return &procFile[i];
+  }
+  return NULL;
+}
+
+/* Create a new container_file structure that corresponds to cont */
+static struct container_file* __paje_create_container_file(const char* cont) {
+
+  struct container_file* res = NULL;
+  if(nb_containers >= nb_allocated_containers) {
+    nb_allocated_containers *= 2 ;
+    nb_allocated_containers=(nb_allocated_containers?nb_allocated_containers:100);
+    procFile = realloc(procFile, sizeof(struct container_file)*nb_allocated_containers);
+    assert(procFile);
+  }
+  assert(nb_containers<nb_allocated_containers);
+
+  res = &procFile[nb_containers];
+  nb_containers++;
+  res->closed = 0;
+  strcpy(res->container_name, cont?cont:"");
+
+  if(paje_flags & GTG_FLAG_PAJE_MULTIPLE_FILES) {
+    /* each container write to a separate file */
+    sprintf (res->file_name, "%s_%s.trace", filename, res->container_name);
+    res->file = fopen (res->file_name, "w");
+    if(!res->file)
+      abort();
+  } else {
+    /* all the containers write to the same file */
+    sprintf (res->file_name, "%s.trace", headFilename);
+    res->file = headFile;
+  }
+  return res;
+}
+
+
+static void __paje_destroy_container(struct container_file* p_cont) {
+  if(p_cont->closed > 0)
+    /* this container is already destroyed */
+    return;
+
+  FLUSH(p_cont->file);
+  fclose (p_cont->file);
+  p_cont->closed = 1;
+}
+
 #if HAVE_LIBZ
 static z_stream z;
 static int compression_ratio = 0;
@@ -122,65 +210,6 @@ trace_return_t pajeSetCompress(int val)
 
 #endif
 
-/* Insert a doubleLinkList_t item in the sorted doubleLinkedList starting with first */
-void insert (doubleLinkList_t** first, doubleLinkList_t* item){
-    varPrec time;
-    doubleLinkList_t* iter = *first;
-    if (!first || !item){
-        fprintf (stderr, "Invalid insert parameter. Leaving\n");
-        exit(-1);
-    }
-    time = item->current.time;
-    /* If nil list, item is the first and leaving */
-    if (iter==NULL){
-        *first=item;
-        item->prev = NULL;
-        item->next = NULL;
-        /*  printf (stderr, "Nil list\n"); */
-        return;
-    }
-    /* If to insert in first pos */
-    if (time<=iter->current.time){
-        item->prev = NULL;
-        item->next = *first;
-        if (item->next)
-            (item->next)->prev = item;
-        *first = item;
-        return;
-    }
-    /* Positionning iterator to insert in a sorted list */
-    while (time>iter->current.time && iter->next)
-        iter = iter->next;
-
-    if (time<=iter->current.time){
-        item->prev = iter->prev;
-        item->next = iter;
-        iter->prev = item;
-        if (item->prev)
-            (item->prev)->next = item;
-    }
-    else{
-        iter->next = item;
-        item->prev = iter;
-        item->next=NULL;
-    }
-}
-
-void initMerge (char* file, int rk, doubleLinkList_t* li){
-    char f[BUFFSIZE];
-
-    sprintf (f, "%s_proc%d.ept", file, rk);
-    li->current.file = fopen (f, "a+");
-    if (!li->current.file){
-        fprintf (stderr, "Failed to open generated trace. Leaving\n");
-        exit (-1);
-    }
-    rewind (li->current.file);
-    li->current.value = NULL;
-    li->current.status = 1;
-}
-
-
 #ifndef HAVE_GETLINE
 ssize_t getline(char** buf, size_t* size, FILE* stream)
 {
@@ -240,90 +269,6 @@ void myGetLine (doubleLinkList_t* li){
         fprintf (stderr, "failed to read\n");
 }
 
-/* need to run 'rm -f' force because there is no file some time*/
-void clean_files (char* fileName, int nbFile){
-    int ret;
-    char cmd[BUFFSIZE];
-    sprintf (cmd, "rm -f %s_root.ept %s_proc*.ept", filename,filename);
-    ret = system(cmd);
-    if (ret == -1) {
-        perror("cleanfiles");
-    }
-}
-
-void merge (char* filename, int nbFile){
-    doubleLinkList_t* first = NULL;
-    doubleLinkList_t* list;
-    doubleLinkList_t* iter;
-    FILE            * res;
-    FILE            * oldF;
-    char            * buf = NULL;
-    char              tmp[BUFFSIZE];
-    char              tmp2[BUFFSIZE];
-    size_t            size;
-    int               i;
-
-    /* Getting the first part of the trace (header+def) */
-#if HAVE_LIBZ
-    if(compression_ratio)
-        sprintf (tmp, "%s.trace.z", filename);
-    else
-#endif
-        sprintf (tmp, "%s.trace", filename);
-
-    res = fopen (tmp, "w+");
-    sprintf (tmp2, "%s_root.ept", filename);
-    oldF = fopen (tmp2,"a+");
-    if (res==NULL || oldF==NULL){
-        fprintf (stderr, "Failed to create the trace file. Leaving\n");
-        exit (-1);
-    }
-    rewind (oldF);
-    do{
-        i = getline (&buf, &size, oldF);
-        if (size==0 || i==-1)
-            break;
-        paje_print (res, "%s", buf);
-    }while (buf[0] != EOF);
-    fclose (oldF);
-    /* Initialising the parallel merge */
-    list = (doubleLinkList_t *)malloc (sizeof (doubleLinkList_t)*nbFile);
-    for (i=0;i<nbFile;i++)
-        initMerge (filename, i, list+i);
-
-    for (i=0;i<nbFile;i++){
-        myGetLine (list+i);
-        insert (&first, list+i);
-    }
-
-    /*  fprintf (stderr, "Nb file = %d \n La liste :\n", nbFile); */
-    /*  for (i=0;i<nbFile;i++)                                     */
-    /*      fprintf (stderr, "L1 = %s\n", list[i].current.value); */
-
-    while (first && first->current.status != 0){
-        paje_print (res, "%s", first->current.value);
-        myGetLine (first);
-        if (first->current.status==0){
-            first=first->next;
-            if (first)
-                first->prev = NULL;
-        }
-        else{
-            iter = first;
-            first = first->next;
-            if (first)
-                first->prev = NULL;
-            if (first && iter->current.status != 0)
-                insert (&first, iter);
-        }
-        /* for (i=0;i<nbFile;i++) */
-        /*     fprintf (stderr, "L1 = %s, st=%d\n", list[i].current.value,list[i].current.status); */
-        /* fprintf (stderr, "<---------------------------->\n\n"); */
-    }
-    fclose (res);
-    clean_files (filename, nbFile);
-}
-
 paje_color_t Paje_get_color(gtg_color_t p_color) {
     /** \todo it is not const*/
     int ret;
@@ -352,55 +297,51 @@ pajeGetName (int procNb){
 
 trace_return_t my_open (int rk){
     trace_return_t ret = TRACE_ERR_OPEN;
-    char f[BUFFSIZE];
 
-    if (!procFile){
-        if(!(paje_flags & GTG_FLAG_USE_MPI)) {
-            procFile = headFile;
-        } else {
-            sprintf (f, "%s_proc%d.ept", filename, rk);
-            procFile = fopen (f, "w");
-        }
-        if (!procFile)
-            return ret;
-        ret = TRACE_SUCCESS;
+    if (!headFile){
+      sprintf (headFilename, "%s_proc%d.ept", filename, rk);
+      headFile = fopen (headFilename, "w");
     }
+    if (!headFile)
+      return ret;
+    ret = TRACE_SUCCESS;
     return ret;
 }
 
 trace_return_t my_close (){
     trace_return_t ret = TRACE_ERR_CLOSE;
-    if (procFile){
-        ret = (trace_return_t) fclose (procFile);
-        procFile = NULL;
+    if (headFile){
+        ret = (trace_return_t) fclose (headFile);
+        headFile = NULL;
     }
     return ret;
 }
 
 trace_return_t pajeInitTrace   (const char* filenam, int rk, gtg_flag_t flags, int fmt) {
-    char file[BUFFSIZE];
     trace_return_t ret = TRACE_ERR_OPEN;
 
     my_close ();
     rank = rk;	/* fixme: why there is need to copy filenam ????*/
     filename = (char *)malloc (sizeof (char)* (strlen (filenam)+1));
     strcpy (filename, filenam);
-    procFile = NULL;
+    headFile = NULL;
 
     paje_flags = flags;
     if( rank==0 ||
-        ! (flags & GTG_FLAG_USE_MPI)) {
-
-        sprintf (file, "%s_root.ept", filename);
-        headFile = fopen (file, "w");
+        (! (flags & GTG_FLAG_USE_MPI) && !(flags & GTG_FLAG_PAJE_MULTIPLE_FILES))) {
+        sprintf (headFilename, "%s_root.ept", filename);
+        headFile = fopen (headFilename, "w");
         if (!headFile){
-            fprintf (stderr, "Failed to open file %s. \n Leaving\n", file);
+            fprintf (stderr, "Failed to open file %s. \n Leaving\n", headFilename);
             return ret;
         }
 
-        pajeInitHeaderData( fmt );
+        pajeInitHeaderData( fmt, flags);
 
         FLUSH(headFile);
+    } else {
+      /* another process writes the header, so we don't have to */
+      _is_paje_header_written = 1;
     }
     return my_open(rank);
 }
@@ -502,40 +443,63 @@ trace_return_t pajeAddEntityValue   (const char* alias, const char* entType,
     return TRACE_ERR_WRITE;
 }
 
-trace_return_t pajeAddContainer(varPrec time, const char* alias,
-                                const char*  type, const char* container,
-                                const char*  name, const char* file){
+trace_return_t pajeAddContainer_generic(varPrec time, const char* alias,
+					const char*  type, const char* container,
+					const char*  name, const char* file){
     PRINT_PAJE_HEADER();
 
     if (headFile){
-        if(container && strcmp(container, "(null)")!=0 )
-            fprintf (headFile, "%d %.13e \"%s\" \"%s\" \"%s\" \"%s\" \"%s\"\n",
-                     paje_eventdefs[GTG_PAJE_EVTDEF_CreateContainer].id, time, name, type, container, alias, file);
-        else
-            fprintf (headFile, "%d %.13e \"%s\" \"%s\" 0 \"%s\" \"%s\"\n",
-                     paje_eventdefs[GTG_PAJE_EVTDEF_CreateContainer].id, time, name, type, alias, file);
+      fprintf (headFile, "%d %.13e \"%s\" \"%s\"",
+	       paje_eventdefs[GTG_PAJE_EVTDEF_CreateContainer].id, time, name, type);
 
-        FLUSH(headFile);
-        return TRACE_SUCCESS;
+      if(container && strcmp(container, "(null)")!=0 ) {
+	fprintf(headFile, " \"%s\"", container);
+      }else{
+	fprintf(headFile, " 0");
+      }
+
+      fprintf(headFile, " \"%s\"", alias);
+
+      if(! null_string(file)) {
+	fprintf(headFile, " \"%s\"", file);
+      }
+
+      fprintf(headFile,"\n");
+      FLUSH(headFile);
+      return TRACE_SUCCESS;
     }
     return TRACE_ERR_WRITE;
 }
 
+
+trace_return_t pajeAddContainer(varPrec time, const char* alias,
+				const char*  type, const char* container,
+				const char*  name, const char* file){
+  PRINT_PAJE_HEADER();
+  if (headFile){
+    if(null_string(file)) {
+      struct container_file* p_cont = __paje_create_container_file(alias);
+      if(paje_flags & GTG_FLAG_PAJE_MULTIPLE_FILES) {
+	return pajeAddContainer_generic(time, alias, type, container, name, p_cont->file_name);
+      }
+    }
+    return pajeAddContainer_generic(time, alias, type, container, name, file);
+  }
+  return TRACE_ERR_WRITE;
+}
+
 trace_return_t pajeSeqAddContainer   (varPrec time, const char* alias    ,
                                       const char*  type, const char* container,
                                       const char*  name){
     PRINT_PAJE_HEADER();
 
     if (headFile){
-        if(container && strcmp(container, "(null)")!=0 )
-            fprintf (headFile, "%d %.13e \"%s\" \"%s\" \"%s\" \"%s\"\n",
-                     paje_eventdefs[GTG_PAJE_EVTDEF_CreateContainer].id, time, name, type, container, alias);
-        else
-            fprintf (headFile, "%d %.13e \"%s\" \"%s\" 0 \"%s\"\n",
-                     paje_eventdefs[GTG_PAJE_EVTDEF_CreateContainer].id, time, name, type, alias);
-
-        FLUSH(headFile);
-        return TRACE_SUCCESS;
+      struct container_file* p_cont = __paje_create_container_file(alias);
+      if(paje_flags & GTG_FLAG_PAJE_MULTIPLE_FILES) {
+	return pajeAddContainer_generic(time, alias, type, container, name, p_cont->file_name);
+      } else {
+	return pajeAddContainer_generic(time, alias, type, container, name, NULL);
+      }
     }
     return TRACE_ERR_WRITE;
 }
@@ -545,38 +509,48 @@ trace_return_t pajeDestroyContainer     (varPrec time, const char*  name,
     PRINT_PAJE_HEADER();
 
     if (headFile){
-        fprintf (headFile, "%d %.13e \"%s\" \"%s\"\n",
-                 paje_eventdefs[GTG_PAJE_EVTDEF_DestroyContainer].id, time, name, type);
+      struct container_file* p_cont = __paje_find_container(name);
 
-        FLUSH(headFile);
-        return TRACE_SUCCESS;
+      fprintf (p_cont->file, "%d %.13e \"%s\" \"%s\"\n",
+	       paje_eventdefs[GTG_PAJE_EVTDEF_DestroyContainer].id, time, name, type);
+
+      FLUSH(p_cont->file);
+      __paje_destroy_container(p_cont);
+
+      return TRACE_SUCCESS;
     }
     return TRACE_ERR_WRITE;
 }
 
 trace_return_t pajeSetState   (varPrec time, const char* type,
                                const char*  cont, const char* val){
+    FILE *f=NULL;
     if(verbose)
         printf("SetState : type %s, container %s, val %s\n", type, cont, val);
-    if (procFile){
-        fprintf (procFile, "%d %.13e \"%s\" \"%s\" \"%s\"\n",
+
+    GET_FILE_DESCRIPTOR(f, cont);
+    if (f){
+        fprintf (f, "%d %.13e \"%s\" \"%s\" \"%s\"\n",
                  paje_eventdefs[GTG_PAJE_EVTDEF_SetState].id, time, type, cont, val);
 
-        FLUSH(procFile);
+        FLUSH(f);
         return TRACE_SUCCESS;
     }
+    printf("f==NULL (cont=%s)\n", cont);
     return TRACE_ERR_WRITE;
 }
 
 trace_return_t pajePushState   (varPrec time, const char* type,
                                 const char*  cont, const char* val){
+    FILE *f=NULL;
     if(verbose)
         printf("PushState : type %s, container %s, val %s\n", type, cont, val);
-    if (procFile){
-        fprintf (procFile, "%d %.13e \"%s\" \"%s\" \"%s\"\n",
+    GET_FILE_DESCRIPTOR(f, cont);
+    if (f){
+        fprintf (f, "%d %.13e \"%s\" \"%s\" \"%s\"\n",
                  paje_eventdefs[GTG_PAJE_EVTDEF_PushState].id, time, type, cont, val);
 
-        FLUSH(procFile);
+        FLUSH(f);
         return TRACE_SUCCESS;
     }
     return TRACE_ERR_WRITE;
@@ -584,13 +558,15 @@ trace_return_t pajePushState   (varPrec time, const char* type,
 
 trace_return_t pajePopState   (varPrec time, const char* type,
                                const char*  cont){
+    FILE *f=NULL;
     if(verbose)
         printf("PopState : type %s, container %s\n", type, cont);
-    if (procFile){
-        fprintf (procFile, "%d %.13e \"%s\" \"%s\"\n",
+    GET_FILE_DESCRIPTOR(f, cont);
+    if (f){
+        fprintf (f, "%d %.13e \"%s\" \"%s\"\n",
                  paje_eventdefs[GTG_PAJE_EVTDEF_PopState].id, time, type, cont);
 
-        FLUSH(procFile);
+        FLUSH(f);
         return TRACE_SUCCESS;
     }
     return TRACE_ERR_WRITE;
@@ -598,13 +574,15 @@ trace_return_t pajePopState   (varPrec time, const char* type,
 
 trace_return_t pajeAddEvent   (varPrec time, const char* type,
                                const char*  cont, const char* val){
+    FILE *f=NULL;
     if(verbose)
         printf("AddEvent : type %s, cont %s, val %s\n", type, cont, val);
-    if (procFile){
-        fprintf (procFile, "%d %.13e \"%s\" \"%s\" \"%s\"\n",
+    GET_FILE_DESCRIPTOR(f, cont);
+    if (f){
+        fprintf (f, "%d %.13e \"%s\" \"%s\" \"%s\"\n",
                  paje_eventdefs[GTG_PAJE_EVTDEF_NewEvent].id, time, type, cont, val);
 
-        FLUSH(procFile);
+        FLUSH(f);
         return TRACE_SUCCESS;
     }
     return TRACE_ERR_WRITE;
@@ -613,14 +591,18 @@ trace_return_t pajeAddEvent   (varPrec time, const char* type,
 trace_return_t pajeStartLink   (varPrec time, const char* type,
                                 const char*   cont, const char* src,
                                 const char*   val , const char* key){
+    FILE *f=NULL;
     if (verbose)
         printf ("Start link: type: %s container: %s src: %s val: %s key %s\n",
                 type, cont, src, val, key);
-    if (procFile){
-        fprintf (procFile, "%d %.13e \"%s\" \"%s\" \"%s\" \"%s\" \"%s\"\n",
+
+    GET_FILE_DESCRIPTOR(f, cont);
+
+    if (f){
+        fprintf (f, "%d %.13e \"%s\" \"%s\" \"%s\" \"%s\" \"%s\"\n",
                  paje_eventdefs[GTG_PAJE_EVTDEF_StartLink].id, time, type, cont, src, val, key);
 
-        FLUSH(procFile);
+        FLUSH(f);
         return TRACE_SUCCESS;
     }
     return TRACE_ERR_WRITE;
@@ -629,14 +611,18 @@ trace_return_t pajeStartLink   (varPrec time, const char* type,
 trace_return_t pajeEndLink   (varPrec time, const char* type,
                               const char*   cont, const char* dest,
                               const char*   val , const char* key){
+    FILE *f=NULL;
     if (verbose)
         printf ("End link type: %s container: %s src: %s val: %s key %s\n",
                 type, cont, dest, val, key);
-    if (procFile){
-        fprintf (procFile, "%d %.13e \"%s\" \"%s\" \"%s\" \"%s\" \"%s\"\n",
+
+    GET_FILE_DESCRIPTOR(f, cont);
+
+    if (f){
+        fprintf (f, "%d %.13e \"%s\" \"%s\" \"%s\" \"%s\" \"%s\"\n",
                  paje_eventdefs[GTG_PAJE_EVTDEF_EndLink].id, time, type, cont, dest, val, key);
 
-        FLUSH(procFile);
+        FLUSH(f);
         return TRACE_SUCCESS;
     }
     return TRACE_ERR_WRITE;
@@ -644,14 +630,18 @@ trace_return_t pajeEndLink   (varPrec time, const char* type,
 
 trace_return_t pajeSetVar   (varPrec time, const char*  type,
                              const char*  cont, varPrec val){
+    FILE *f=NULL;
     if (verbose)
         printf ("SetVar: type: %s container: %s value: %e\n",
                 type, cont, val);
-    if (procFile){
-        fprintf (procFile, "%d %.13e \"%s\" \"%s\" %e\n",
+
+    GET_FILE_DESCRIPTOR(f, cont);
+
+    if (f){
+        fprintf (f, "%d %.13e \"%s\" \"%s\" %e\n",
                  paje_eventdefs[GTG_PAJE_EVTDEF_SetVariable].id, time, type, cont, val);
 
-        FLUSH(procFile);
+        FLUSH(f);
         return TRACE_SUCCESS;
     }
     return TRACE_ERR_WRITE;
@@ -659,14 +649,18 @@ trace_return_t pajeSetVar   (varPrec time, const char*  type,
 
 trace_return_t pajeAddVar   (varPrec time, const char*  type,
                              const char*  cont, varPrec val){
+    FILE *f=NULL;
     if (verbose)
         printf ("AddVar: type: %s container: %s value: %e\n",
                 type, cont, val);
-    if (procFile){
-        fprintf (procFile, "%d %.13e \"%s\" \"%s\" %e\n",
+
+    GET_FILE_DESCRIPTOR(f, cont);
+
+    if (f){
+        fprintf (f, "%d %.13e \"%s\" \"%s\" %e\n",
                  paje_eventdefs[GTG_PAJE_EVTDEF_AddVariable].id, time, type, cont, val);
 
-        FLUSH(procFile);
+        FLUSH(f);
         return TRACE_SUCCESS;
     }
     return TRACE_ERR_WRITE;
@@ -674,14 +668,16 @@ trace_return_t pajeAddVar   (varPrec time, const char*  type,
 
 trace_return_t pajeSubVar   (varPrec time, const char*  type,
                              const char*  cont, varPrec val){
+    FILE *f=NULL;
     if (verbose)
         printf ("SubVar: type: %s container: %s value: %e\n",
                 type, cont, val);
-    if (procFile){
-        fprintf (procFile, "%d %.13e \"%s\" \"%s\" %e\n",
+    GET_FILE_DESCRIPTOR(f, cont);
+    if (f){
+        fprintf (f, "%d %.13e \"%s\" \"%s\" %e\n",
                  paje_eventdefs[GTG_PAJE_EVTDEF_SubVariable].id, time, type, cont, val);
 
-        FLUSH(procFile);
+        FLUSH(f);
         return TRACE_SUCCESS;
     }
     return TRACE_ERR_WRITE;
@@ -691,56 +687,250 @@ trace_return_t pajeAddComment   (const char*  comment){
     if (verbose)
         printf ("AddComment: %s\n",comment);
 
-    if (procFile){
-        fprintf(procFile,"#%s\r\n",comment);
-        FLUSH(procFile);
+    if (headFile){
+        fprintf(headFile,"#%s\r\n",comment);
+        FLUSH(headFile);
         return TRACE_SUCCESS;
     }
     return TRACE_ERR_WRITE;
 }
 
+int acquire_lock(const char* lockfile) {
+  int lock_fd, rc;
+  lock_fd = open(lockfile, O_RDWR | O_CREAT, 0666);
+  if(lock_fd < 0) {
+    printf("open(%s) returned %d (%s)\n", lockfile, errno, strerror(errno));
+    abort();
+  }
+
+  rc = flock(lock_fd, LOCK_EX);
+  if(rc != 0) {
+    printf("flock returned %d\n", rc);
+    abort();
+  }
+  return lock_fd;
+}
+
+void release_lock(const char* lockfile, int lock_fd){
+  int rc = flock(lock_fd, LOCK_UN);
+  if(rc !=0){
+    printf("unlock returned %d\n", rc);
+    abort();
+  }
+  /* todo: delete the lock file */
+}
+
+#define MAX_LINE_SIZE 1024
+struct paje_line_t{
+  char *line;
+  int len;			/* length of the line */
+  size_t size;			/* allocated size of line */
+  int code;
+  double timestamp;
+  int eof;
+};
+
+static void __init_paje_line(struct paje_line_t*l) {
+  l->size=MAX_LINE_SIZE;
+  l->line=malloc(sizeof(char)*l->size);
+  l->len=0;
+  l->code=-1;
+  l->timestamp=-1;
+  l->eof=0;
+}
+
+void read_line(FILE*f, struct paje_line_t *l) {
+  int rc;
+  l->len = getline(&l->line, &l->size, f);
+  if(l->len<0) {
+    if(!feof(f)) {
+      printf("getline failed!\n");
+      abort();
+    }
+    l->eof=1;
+    return;
+  }
+  /* getline managed to read something */
+  rc = sscanf(l->line, "%d %lf", &l->code, &l->timestamp);
+  if(rc != 2) {
+    /* this is probably because the line starts with % */
+    l->code=-1;
+    l->timestamp=-1;
+  }
+}
+
+void write_line(FILE*f, struct paje_line_t*l) {
+  fprintf(f, "%s", l->line);
+}
+
+/* check if input1 (resp. input2) is empty and if so, rename input2 (resp.
+ * input1) into output)
+ * return 0 if none of the files were renamed
+ */
+static int __try_to_rename(const char *input1,
+			   const char *input2,
+			   const char *output) {
+  /* check if input1 exists */
+  struct stat st;
+  int rc = stat(input1, &st);
+  if(rc <0 && errno == ENOENT) {
+    /* input1 does not exist. we can simply rename input2 to output */
+    rc = rename(input2, output);
+    assert(rc == 0);
+    return 1;
+  }
+
+  /* check if input2 exists */
+  rc = stat(input2, &st);
+  if(rc <0 && errno == ENOENT) {
+    /* input2 does not exist. we can simply rename input1 to output */
+    rc = rename(input1, output);
+    assert(rc == 0);
+    return 1;
+  }
+
+  /* both files exist */
+  return 0;
+}
+
+/* merge input1 and input2 into output */
+static void __merge_files(const char *input1,
+			  const char *input2,
+			  const char *output) {
+  FILE* input1_file;
+  FILE* input2_file;
+  FILE* output_file;
+  struct paje_line_t line_input, line_merge;
+
+  if(__try_to_rename(input1, input2, output)) {
+    goto out;
+  }
+
+  /* both input1 and input2 exist */
+
+  /* open the merge file */
+  input1_file = fopen(input1, "r");
+  assert(input1_file);
+
+  /* open the input file */
+  input2_file = fopen(input2, "r");
+  assert(input2_file);
+
+  /* open the output file */
+  output_file = fopen(output, "w");
+  assert(output_file);
+
+  /* merge input2_file and input1_file */
+  __init_paje_line(&line_input);
+  __init_paje_line(&line_merge);
+  read_line(input2_file, &line_input);
+  read_line(input1_file, &line_merge);
+
+  /* merge input and merge files */
+  while(!line_input.eof && ! line_merge.eof) {
+    if(line_input.timestamp<line_merge.timestamp) {
+      write_line(output_file, &line_input);
+      read_line(input2_file, &line_input);
+    } else {
+      write_line(output_file, &line_merge);
+      read_line(input1_file, &line_merge);
+    }
+  }
+
+  /* copy the remaining of the files */
+  while(!line_input.eof) {
+    write_line(output_file, &line_input);
+    read_line(input2_file, &line_input);
+  }
+
+  while(!line_merge.eof) {
+    write_line(output_file, &line_merge);
+    read_line(input1_file, &line_merge);
+  }
+
+  /* close files and delete merge and input */
+  fclose(output_file);
+  fclose(input2_file);
+  fclose(input1_file);
+  unlink(input1);
+  unlink(input2);
+
+ out:
+  return;
+}
+
+/* merge headFilename to the other headFile (from other processes) into a file named
+ * <base_filename>.trace
+ */
+static void merge_contrib(const char *base_filename) {
+  int lock_fd;
+  int rc;
+  char merged_filename[BUFFSIZE]; /* the other file to merge */
+  char output_filename[BUFFSIZE]; /* the resulting file */
+  char lockfile[BUFFSIZE];
+
+  /* <filename>.trace will contain the merge of <filename>.ept and <filename>.trace.merged
+   */
+
+  /* todo: check if compression still works */
+#if HAVE_LIBZ
+  if(compression_ratio)
+    sprintf (output_filename, "%s.trace.z", base_filename);
+  else
+#endif
+    sprintf (output_filename, "%s.trace", base_filename);
+  sprintf (merged_filename, "%s.trace.merged", base_filename);
+  sprintf (lockfile, "%s.trace.lock", base_filename);
+
+  /* make sure we have exclusive access to files */
+  lock_fd = acquire_lock(lockfile);
+  assert(lock_fd>=0);
+  {
+    /* Currently, <filename>.trace contains the merge of the other *.ept files,
+     * let's rename it to <filename>.trace.merge
+     */
+    rc = rename(output_filename, merged_filename);
+    if(rc != 0 && errno != ENOENT) {
+      /* rename failed */
+      fprintf(stderr, "gtg: failed to rename %s into %s\n", output_filename, merged_filename);
+      abort();
+    }
+
+    __merge_files(merged_filename, headFilename, output_filename);
+  }
+  release_lock(lockfile, lock_fd);
+}
+
+static trace_return_t
+__pajeEndTrace_generic (int should_merge) {
+  int i;
+  my_close ();
+  /* Wait for all proc to finish writing their trace */
+  if(paje_flags & GTG_FLAG_PAJE_MULTIPLE_FILES) {
+    for(i=0;i<nb_containers; i++) {
+      __paje_destroy_container(&procFile[i]);
+    }
+  }
+
+  if(should_merge) {
+    merge_contrib(filename);
+  }
+
+  if (nameTmp)
+    free (nameTmp);
+  nameTmp = NULL;
+  if (filename)
+    free (filename);
+  filename = NULL;
+  return TRACE_SUCCESS;
+}
 
 trace_return_t pajeEndTrace (){
-    int size = 1;
-    my_close ();
-    /* Wait for all proc to finish writing their trace */
-#ifdef USE_MPI
-    if(paje_flags & GTG_FLAG_USE_MPI) {
-        MPI_Barrier (MPI_COMM_WORLD);
-        MPI_Comm_size (MPI_COMM_WORLD, &size);
-        if(verbose)
-            fprintf (stderr, "USING MPI\n");
-    }
-    if (rank==0)
-#endif	/* USE_MPI */
-        merge (filename, size);
-
-    if (nameTmp)
-        free (nameTmp);
-    nameTmp = NULL;
-    if (filename)
-        free (filename);
-    filename = NULL;
-    return TRACE_SUCCESS;
+  return __pajeEndTrace_generic(1);
 }
 
 trace_return_t viteEndTrace (){
-    my_close ();
-    if (filename)
-        free (filename);
-    if (nameTmp)
-        free (nameTmp);
-    filename = NULL;
-    nameTmp = NULL;
-    return TRACE_SUCCESS;
-}
-
-
-/*
- * Hack that will be removed
- */
-FILE *pajeGetProcFile() {
-    return procFile;
+  return __pajeEndTrace_generic(0);
 }
 
 #endif
